module PCResolution(
  binFactor,binResolvent,
  allBinResolvents,
  istautology,
  resolutionWithSupport,
  subsumes,
  subsumptionDel,
  strictsubsumptionDel,
  search,
  filterid, totalFilter, 
  dfs, dfstrace, dfstracelen,
  bfs, bfstrace, bfstracelen,
  rmrmfact, rmrm,
  fewestLitPref, 
  bestFirstSearch,
  tightStrategy)
where

import PCFormula
import ToClause
import List
import Maybe
import Debug.Trace
import System

foreachSplitAt xs = [(splitAt i xs) | i<-[0..length xs-1]] 
                                   

allsplit4 xs = [(l,head r,lr,head rr, tail rr) | 
                     (l,r) <- foreachSplitAt xs, 
                     (lr,rr)<-foreachSplitAt (tail r)]


binFactor (Lit p) (Lit q)
   | p == q = True
   | otherwise = False
binFactor (NegLit p) (NegLit q) 
   | p == q = True
   | otherwise = False
binFactor _ _ = False



binResolvent (Lit p) (NegLit q) 
   | p == q = True
   | otherwise = False
binResolvent (NegLit p) (Lit q) 
   | p == q = True
   | otherwise = False
binResolvent x y = False

tautologyFactors (Lit p) (NegLit q)
   | p == q = True
   | otherwise = False
tautologyFactors (NegLit p) (Lit q)
   | p == q = True
   | otherwise = False
tautologyFactors  _ _ = False

factor = removedup

factorAll xs = map factor xs

allBinResolvents x y = [lx++(tail rx)++ly++(tail ry) |
                         (lx,rx)<-foreachSplitAt x, 
                         (ly,ry)<-foreachSplitAt y,
                         binResolvent (head rx) (head ry)]

istautology xs = any tautologyHead (map snd (foreachSplitAt xs))
     where tautologyHead (x:xs) = any (tautologyFactors x) xs 
  
removeTautology  xs = filter (not . istautology) xs 


resolveAlltoOne clauses supclause =
       concat  (map (allBinResolvents supclause) clauses)


resolutionWithSupport support clauses = factorAll 
					  (concatMap (resolveWithOne clauses)
                                                  (foreachSplitAt support))


resolveWithOne clauses (l,r) = resolveAlltoOne (l++(tail r)++clauses) (head r)

setdiff xs ys =  foldl (\ l x -> filter (x/=) l) xs ys

removedup xs =  foldr (\ x l-> x:(delete x l)) [] xs

--removedup = nub

subsumes x y = null(setdiff x y)
subsumptionDel xs = nubBy subsumes xs

strictsubsumptionDel xs =
  let 
    ys = subsumptionDel xs
  in (reverse . subsumptionDel .reverse) ys

search filtermap addTo support clauses =
  let 
    new =  filtermap (resolutionWithSupport support clauses) support
  in if null new
     then False
     else if (any null new)
     then True
     else let
            newsup = strictsubsumptionDel (addTo new support)
          in if (fixpoint newsup support)
             then False
             else search filtermap addTo newsup clauses

filterid new old = new

totalFilter new old = 
    let
      d = setdiff (removeTautology (map factor new)) old
    in --trace ("new:\n"++
       --      (clauseSetToString new)++"\nold:\n"++
       --      (clauseSetToString old)++"\ndiff:\n"++ 
       --      (clauseSetToString d))
             d

dfs new old = new++old
dfstrace new old = (trace ("new set: "++show new) new)++old
dfstracelen new old = (trace (show (length new)) new)++old

bfs new old = old++new
bfstrace new old = old++(trace ("new set: "++show new) new)
bfstracelen new old = old++(trace (show (length new)) new)

rmrmfact = removedup . removeTautology . (\ xs -> map factor xs)

rmrm = removedup . removeTautology


numLitCmp xs ys = 
    let 
     nxs = length xs
     nys = length ys
    in 
      if  nxs < nys 
      then LT
      else if  nxs == nys
      then EQ
      else GT


fewestLitPref xs = sortBy numLitCmp xs

bestFirstSearch traceon strategy support clauses =
  let 
    newsup =  strategy (resolutionWithSupport support clauses) support
  in if fixpoint newsup support
     then False
     else if (any null newsup)
     then True
     else bestFirstSearch traceon 
                          strategy 
                          (proofTrace traceon newsup)
                          clauses



tightStrategy new support =
    let 
      cleannew = removeTautology (map factor new)
    in 
      (subsumptionDel . fewestLitPref) (cleannew++support)
        

fixpoint newset oldset = 
    all (\ xnew -> any (\xold ->  subsumes xold xnew) oldset) newset



proofTrace 0 clauses = clauses
proofTrace 1 clauses = trace (show (length clauses)) clauses 
proofTrace 2 clauses = 
    trace (clauseSetToString clauses)
          clauses



